﻿using System;
using System.Collections.Generic;
using System.Text;

namespace _11_GenerischeListeMitSuche
{
    #region hide
    // : IComparable<Liste<T>> where T : IComparable<T>
    #endregion
    class Liste<T> where T:IComparable<T>
    {
        class LItem
        {
            public T data;
            public LItem next = null, prev = null;
            public LItem(T Data) { data = Data; }

            public override string ToString()
            => $"Name: {data}, prev: {(prev == null ? "null" : prev.data.ToString())}, next: {(next == null ? "null" : next.data.ToString())}";
        }

        LItem first = null, last = null;
        public int Count { get; private set; }
        public Liste() { Count = 0; }
        public int CompareTo(Liste<T> other) => Count - other.Count;

        public void AddEnd(T Data)
        {
            LItem newItem = new LItem(Data);
            Count++;
            if (first == null)
                first = last = newItem;
            else  // Es gibt schon Listenelemente
            {
                last.next = newItem;
                newItem.prev = last;
                last = newItem;
            }
        }
        public void AddFront(T Data)
        {
            LItem newItem = new LItem(Data);
            Count++;
            if (first == null)
                first = last = newItem;
            else
            {
                newItem.next = first;
                first.prev = newItem;
                first = newItem;
            }
        }
        public void AddSorted(T InsData)
        {
            // 1. Fall: Leere Liste oder Anfügen am Listenende
            if (first == null || last.data.CompareTo(InsData) <= 0)
                AddEnd(InsData);
            else if (InsData.CompareTo(first.data) <= 0)
                // 2. Fall: Anfügen am Listenende
                AddFront(InsData);
            else
            {
                // Wir wissen: Das neue Element ist NICHT das erste und nicht das letzte Element
                LItem newItem = new LItem(InsData);
                Count++;

                LItem item = first;
                while (item.next.data.CompareTo(InsData) < 0)
                    item = item.next;
                // bis hierher identisch zur einfach-verketteten Liste
                newItem.next = item.next; // Vorwärtsverkettung vom neuen Element zur Restliste
                newItem.prev = item;      // Rückwärtsverkettung vom neuen Elenent zum Listenanfangsstück
                item.next.prev = newItem; // Rückwärtsverkettung vom nachfolgenden Listenelement
                item.next = newItem;      // Vorwärtsverkettung zum neuen Element
            }
        }
        public void DeleteFirst()
        {
            // Fall 1: Liste ist leer
            // Fall 2: Liste besteht nur aus EINEM Element
            // Fall 3: Liste hat mehr als ein Element

            if (first == null)  // Fall 1
                return;
            Count--;
            if (first == last)  // Fall 2
                first = last = null;
            else
            {
                first = first.next;
                first.prev = null;
            }
        }
        public void DeleteLast()
        {
            // Fall 1: Liste ist leer
            // Fall 2: Liste besteht nur aus EINEM Element
            // Fall 3: Liste hat mehr als ein Element

            if (first == null)  // Fall 1
                return;
            Count--;
            if (first == last)  // Fall 2
                first = last = null;
            else
            {
                last = last.prev;
                last.next = null;
            }
        }
        public void DeleteByName(T DelData)
        {
            // Fall 1: Liste ist leer
            // Fall 2: Das gesuchte Element befindet sich am Anfang
            // Fall 3: Das gesuchte Element befindet sich am Ende
            // Fall 4: Das gesuchte Element befindet sich mittendrin

            if (first == null)  // Fall 1
                return;
            if (first.data.CompareTo(DelData) == 0)   // Fall 2
                DeleteFirst();
            else if (last.data.CompareTo(DelData) == 0) // Fall 3
                DeleteLast();
            else
            {
                // Fall 4: Wir wissen:
                // Das zu löschende Element befindet sich weder am Anfang noch am Ende
                //    --> Es gibt ein Vorgänger- UND ein Nachfolger-Element
                Count--;
                LItem item = first.next;
                while (item.next != null && item.data.CompareTo(DelData) != 0)
                    item = item.next;
                if (item.next != null)
                {
                    // d.h. Element wurde gefunden und item zeigt auf dieses Element
                    //                                  20 - 21 - 22
                    item.prev.next = item.next;      // 20 -> 22 Nachfolger
                    item.next.prev = item.prev;      // 22 -> 20 Vorgänger
                }
            }
        }
        public void Print()
        {
            for (LItem item = first; item != null; item = item.next)
            {
                Console.WriteLine(item.data);
            }
            Console.WriteLine("-------------------");
        }
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            for (LItem item = first; item != null; item = item.next)
            {
                sb.Append(item.data.ToString());
                sb.Append("  ");
            }
            return sb.ToString();
        }
        public void PrintReverse()
        {
            for (LItem item = last; item != null; item = item.prev)
            {
                Console.WriteLine(item.data);
            }
            Console.WriteLine("-------------------");
        }
        public bool Exist(T findData)
        {
            for (LItem item = first; item != null; item = item.next)
            {
                if (item.data.CompareTo(findData) == 0)
                    return true;
            }
            return false;
        }
        public IEnumerable<T> Reverse()
        {
            for (LItem item = last; item != null; item = item.prev)
                yield return item.data;
        }
        public IEnumerator<T> GetEnumerator()
        {
            for (LItem item = first; item != null; item = item.next)
                yield return item.data;
        }
        #region hide
        //public delegate bool MeinFilter<T>(T var);
        //public delegate bool Predicate<T>(T var);  Bereits von Microsoft vordefiniert
        // Predicate<T> filterFunktion
        #endregion

        public IEnumerable<T> Filter(Predicate<T> filterFunktion)
        {
            for (LItem item = first; item != null; item = item.next)
                if (filterFunktion(item.data))  // Filterfunktion 
                    yield return item.data;
        }
    }
}

